<?php

require_once("tcpdf/tcpdf.php");
require_once("tcpdf/fpdi.php");

/*
 * The PDFInvoice class is used for generation invoices in PDF format. You cam modify this code
 * to get a custom view for your invoices. To generate PDF documents is used TCPDF library.
 * Please refer to http://www.tcpdf.org/ to learn how to modify PDF generation code. Please pay
 * attention that your modifications will affect immediately invoices that is sent to your
 * customers and is printed for you.
 */
class PDFInvoice extends FPDI
{
	/*
	 * Constructor
	 *
	 * The $args array are:
	 *   invoice => object of Invoice type (mandatory)
	 *   issuerCompany => object of Company type (mandatory)
	 *   configParams => configuration parameters from 'Invoice Settings' page in format 'name' => 'value' (mandatory)
	 *     ("invoice_title", "header_align", "measurement_unit", "horizontal_size", "vertical_size", "horizontal_indent",
   	 *      "vertical_indent", "font_size", "tax_id_title", "header_align", "tax_included_title", "sale_terms", "custom_footer"
   	 *		"invoice_logo", "show_invoice_logo")
   	 *   $transactions // array of Transaction objects (mandatory)
   	 *   $issuerLocale // issuer locale (string) (mandatory)
   	 *   $userLocale // invoice user locale. (string) (mandatory)
	 *   stationeryPdfFile => path to stationery PDF file (optional)
	 *   certificate => object of Certificate type (optional)
	 *
	 * All objects have properties that is equal to database fields:
	 *   $invoiceNumber = $invoice->invoice_number; // number of invoice
	 * To get access to database-related objects is used the following syntax:
	 *   $client = $invoice->Client; // customer
	 *   $invoiceLineItems = $invoice->InvoiceLineItemList(); // array of InvoiceLineItem objects
 	 *   foreach ($invoiceLineItems as $invoiceLineItem) {
	 *     if (0 == $invoiceLineItem->package_addon_id) {
	 *       // $invoiceLineItem is subscription item
	 *       $subscription = $invoiceLineItem->Package; // subscription, object of Package type
	 *       $subscription_attributes = $subscription->PackageAttributesList(); // array of PackageAttributes objects
	 *     } else {
	 *       // $invoiceLineItem is subscription addon item
	 *       $subscription_addons = $invoiceLineItem->Package->PackageAddonList() // array of PackageAddon objects
   	 *     }
	 *   }
	 * Each Transaction object may contain BillingAccount object
	 * 	 $transaction->BillingAccount //object of ClientBillingAccount type (if property 'billing_account_id' > 0)
	 * Access to customer attributes:
	 *   $attribute_id = 4;
	 *   $attribule_value = $invoice->Client->ClientAttributes[$attribute_id]->client_attribute_value;
	 *
	 * @param array $args
	 */
	public function __construct(array $args)
	{
		parent::__construct(PDF_PAGE_ORIENTATION, PDF_UNIT, PDF_PAGE_FORMAT, true, "UTF-8", false);

		// Save mandatory arguments
		$this->invoice = $args["invoice"];
		$this->issuerCompany = $args["issuerCompany"];
		$this->configParams = $args["configParams"];
		$this->transactions = $args["transactions"];
		$this->issuerLocale = $args["issuerLocale"];
		$this->userLocale = $args["userLocale"];
		$this->client = $this->invoice->Client;
		$this->primaryContact = $this->client->PrimaryContact;
		$this->payment_types = Transaction::payment_types();
		$this->initAddressOptions();

		// set default document layout properties
		$this->SetMargins(PDF_MARGIN_LEFT, PDF_MARGIN_TOP, PDF_MARGIN_RIGHT);
		$this->SetFooterMargin(PDF_MARGIN_FOOTER);
		$this->setImageScale(PDF_IMAGE_SCALE_RATIO);
		$this->setLanguageArray(array("w_page" => blmsg("TRANS_PAGE", array(), $this->userLocale)));
		$this->SetDefaultMonospacedFont(PDF_FONT_MONOSPACED);
		$this->SetFont($this->getDefaultFont(), "", $this->getDefaultFontSize());
		$this->setHeaderFont(array($this->getDefaultFont(), "", $this->getDefaultFontSize() + 1));
		$this->setFooterFont(array($this->getDefaultFont(), "", $this->getDefaultFontSize() + 1));

		// init document
		if (isset($args["stationeryPdfFile"])) {
			$stationeryPdfFile = $args["stationeryPdfFile"];
			$page_num = $this->setSourceFile($stationeryPdfFile);
			if (0 < $page_num) {
				$this->stationery = $this->importPage(1);
			}
			$this->AddPage();
			if (1 < $page_num) {
				$this->stationery = $this->importPage(2);
			}
		} else {
			$this->AddPage();
		}

		// calculate auxiliary layout properties
		$dim = $this->getPageDimensions();
		$this->pw = $dim["wk"] - $dim["lm"] - $dim["rm"];  // page width without margins
		$this->ph = $dim["hk"] - $dim["tm"] - $dim["bm"];  // page height without margins
		$this->lm = $dim["lm"];  // page left margin
		$this->tm = $dim["tm"];  // page top margin

		// create fake pdf document for height estimations
		$this->fakePdf = new FPDI(PDF_PAGE_ORIENTATION, PDF_UNIT, PDF_PAGE_FORMAT, true, "UTF-8", false);
		$this->fakePdf->SetMargins(PDF_MARGIN_LEFT, PDF_MARGIN_TOP, PDF_MARGIN_RIGHT);
		$this->fakePdf->SetFont($this->getDefaultFont(), "", $this->getDefaultFontSize());
		$this->fakePdf->AddPage();

		// set document signature
		if (isset($args["certificate"])) {
			$certificate = $args["certificate"];
			$this->setSignature($certificate->certificate, $certificate->private_key, "", $certificate->CAcertificate, 2, array());
		}

		// set document properties
		$this->SetAuthor($this->issuerCompany->company_name);
		$this->SetTitle("Invoice");
		$this->SetSubject("Invoice");

		// add content
		$this->SetFont($this->getDefaultFont(), "", 10);

		$this->addLogo();
		$this->addSenderInfo();
		$this->addCustomerInfo();

		$maxY = $this->GetY();

		$this->addCustomerAdditionalInfo();
		$this->addInvoiceData();
		$this->addSenderAdditionalInfo();

		if ($this->GetY() < $maxY)
			$this->SetY($maxY);
		$this->SetFont($this->getDefaultFont(), "", $this->getDefaultFontSize());

		$this->addInvoiceTitle();
		$this->addBillingPoint();
		$this->addItems();

		$this->fakePdf->setY(0);
		$this->addTotals($this->fakePdf);
		if (($this->getY() + $this->fakePdf->getY()) > ($this->tm + $this->ph)) {
			// element does not fit to page
			$this->AddPage();
		}
		$this->addTotals($this);
		$this->SetY($this->GetY() + 5);
		$this->addPayments();
		$this->addSaleTerms($this);
		$this->addCustomFooter($this);
	}

    /**
	 * Overwritten TCPDF::Header() method. Whenever a page break condition is met, the method is called,
	 * and the break is issued or not depending on the returned value.
	 */
	public function Header()
	{
		if (!is_null($this->stationery)) {
			$this->useTemplate($this->stationery, 0, 0, 0, 0, true);
		}
	}

	private function getDefaultFont()
	{
		static $defaultFont = null;

		if ( !is_null($defaultFont) ) {
			return $defaultFont;
		}

		$localeElements = locale_parse(BillingLocale::getLocale());
		switch ( $localeElements["language"] ) {
			case "ja":
				return ($defaultFont = "arialunicid0-ja");
			case "ko":
				return ($defaultFont = "arialunicid0-ko");
			case "zh":
				switch ( $localeElements["script"] ) {
					case "Hant":
						return ($defaultFont = "arialunicid0-zh_hant");
					default:
						return ($defaultFont = "arialunicid0-zh_hans");
				}
			default:
				return ($defaultFont = "freesans");
		}
	}

	private function getDefaultFontSize()
	{
		return 7;
	}

	private function addLogo()
	{
		if ( 0 == $this->configParams["invoice_show_logo"] ) {
			return;
		}
		$type = strtoupper(substr($this->configParams["invoice_logo"], strrpos($this->configParams["invoice_logo"],"." )+1));
		// Image($file, $x='', $y='', $w=0, $h=0, $type='', $link='', $align='', $resize=false, $dpi=300, $palign='', $ismask=false, $imgmask=false, $border=0, $fitbox=false, $hidden=false, $fitonpage=false)
		// Image best fit is 113x85
        $logoFilename = Config()->invoiceLogoDirectory."/".$this->configParams["invoice_logo"];
        if (file_exists($logoFilename)) {
		    $this->Image( $logoFilename, 20, 20, 0, 0, $type, '', '', false, 150, '', false, false, 1, false, false, false);
        }
	}


	private function addSenderInfo()
	{
		$address = $this->issuerCompany->company_name . "<br />"
			. $this->issuerCompany->company_address_1 . "<br />"
			.(("" != $this->issuerCompany->company_address_2) ? " - " . $this->issuerCompany->company_address_2 : "") . "<br />"
			. $this->issuerCompany->company_city
			. (("" != $this->issuerCompany->company_state || "" != $this->issuerCompany->company_zip) ? ", " . $this->issuerCompany->company_state . " " . $this->issuerCompany->company_zip . "<br />" : "")
			. blmsg("TRANS_ISO_3166_".$this->issuerCompany->Country->countries_iso_2, array(), $this->userLocale);

		$this->writeHTMLCell(114.3, 0, 85.0, 20.0, $address, 0, 1, 0, true);
	}

	private function addCustomerInfo()
	{
		$address = "";
		if (("" != $this->primaryContact->contact_company)) {
			$address .= $this->primaryContact->contact_company . "<br />";
		}
		$address .= $this->primaryContact->contact_fullname . "<br />";
		$address .= $this->primaryContact->contact_address_1 . "<br />"
		. (("" != $this->primaryContact->contact_address_2) ? " - " . $this->primaryContact->contact_address_2 : "") . "<br />"
		. $this->primaryContact->contact_city
		. (("" != $this->primaryContact->contact_state || "" != $this->primaryContact->contact_zip) ? ", " . $this->primaryContact->contact_state . " " . $this->primaryContact->contact_zip : "") . "<br />"
		. blmsg("TRANS_ISO_3166_".$this->primaryContact->Country->countries_iso_2, array(), $this->userLocale);

		$this->setFontSize($this->address_font_size);
		$x = ("left" != $this->configParams["header_align"]) ? $this->address_x : 141 ;
		$this->writeHTMLCell($this->address_w, $this->address_h, $x, $this->address_y, $address, 0, 1, 0, true);
		$this->SetY($this->GetY() + 3);
	}

	private function addInvoiceData()
	{
		$line_items_ids = array();
		// get longest cycle (minimal contract term)
		$this->min_contract_term = "";
		$next_billing_date = 0;
		foreach ($this->invoice->InvoiceLineItemList(array("package_addon_id" => 0)) as $invoiceLineItem) {
			$line_items_ids[] = $invoiceLineItem->package_id;
			if ($next_billing_date < $invoiceLineItem->line_item_date_next_renewal) {
				$next_billing_date = $invoiceLineItem->line_item_date_next_renewal;
				$this->min_contract_term = blmsg($invoiceLineItem->Cycle->cycle_display_name, array(), $this->userLocale);
			}
		}
		$invoice = blmsg("TRANS_INVOICE_PDF_DATE_ENTERED", array(), $this->userLocale) . " : " .  $this->formatValue("date", $this->invoice->invoice_date_entered) . "<br />"
			. blmsg("TRANS_INVOICE_PDF_DATE_DUE", array(), $this->userLocale) . " : " .  $this->formatValue("date", $this->invoice->invoice_date_due) . "<br />"
			. blmsg("TRANS_INVOICE_PDF_NUMBER", array(), $this->userLocale) . " : " .  $this->invoice->invoice_number . "<br />"
			. blmsg("TRANS_INVOICE_PDF_CUSTOMER_NUMBER", array(), $this->userLocale) . " : " .  $this->invoice->client_id . "<br />"
			. blmsg("TRANS_INVOICE_PDF_SUBSCRIPTION_NUMBER", array(), $this->userLocale) . " : " .  implode(",", $line_items_ids) . "<br />";

		$x = ("left" != $this->configParams["header_align"]) ? 130 : $this->address_x ;
		$this->writeHTMLCell($this->address_w, $this->address_h, $x, $this->address_y, $invoice, 0, 1, 0, true);
		$this->SetY($this->GetY() + 3);
	}

	private function addSenderAdditionalInfo()
	{
		$tax_id_title = $this->configParams["tax_id_title"];
		$tax_id_title = ("" != $tax_id_title) ? $tax_id_title : blmsg("TRANS_TAX_ID_NUMBER", array(), $this->userLocale);

		$sender_add_info = (("" != $this->issuerCompany->company_phone1) ? blmsg("TRANS_PRIMARY_PHONE", array(), $this->userLocale) . ": " . $this->issuerCompany->company_phone1 : "") . "<br />"
			. (("" != $this->issuerCompany->company_phone2) ? blmsg("TRANS_SECONDARY_PHONE", array(), $this->userLocale) . ": " . $this->issuerCompany->company_phone2 : "") . "<br />"
			. (("" != $this->issuerCompany->company_tax_id_number) ? $tax_id_title . ": " . $this->issuerCompany->company_tax_id_number : ""). "<br />"
			. (("" != $this->issuerCompany->company_email) ? $this->issuerCompany->company_email : "");

		$x = ("left" != $this->configParams["header_align"]) ? 130 : $this->address_x ;
		$this->writeHTMLCell($this->address_w, $this->address_h, $x, $this->GetY(), $sender_add_info, 0, 1, 0, true);
	}

	private function addCustomerAdditionalInfo()
	{
		$tax_id_title = $this->configParams["tax_id_title"];
		$tax_id_title = ("" != $tax_id_title) ? $tax_id_title : blmsg("TRANS_TAX_ID_NUMBER", array(), $this->userLocale);

		$account_info = (("" != $this->client->tax_id_number) ? $tax_id_title . ": " . $this->client->tax_id_number . "<br />" : "");

		$x = ("left" != $this->configParams["header_align"]) ? $this->address_x : 141 ;
		$this->writeHTMLCell($this->address_w, $this->address_h, $x, $this->GetY(), $account_info, 0, 1, 0, true);
	}

	private function addInvoiceTitle()
	{
		$fs = $this->address_font_size;

		$this->setFontSize($this->address_font_size);
		if (Invoice::STATUS_CREDIT_NOTE == $this->invoice->invoice_status) {
			$invoice_title = blmsg("TRANS_REVERSAL_INVOICE", array(), $this->userLocale);
		} else {
			$invoice_title = $this->configParams["invoice_title"];
			$invoice_title = ("" != $invoice_title) ? $invoice_title : blmsg("TRANS_INVOICE", array(), $this->userLocale);
		}

		$header = "<b>$invoice_title</b><br />";
		$this->writeHTMLCell(0, 0, 20, $this->GetY(), $header, 0, 1, 0, true);
	}

	private function addBillingPoint()
	{
		$billing_point = blmsg("TRANS_BILLING_POINT", array(), $this->userLocale).": ".$this->formatValue("date", $this->invoice->invoice_date_entered)."<br />"
						. blmsg("TRANS_MINIMAL_CONTRACT_TERM", array(), $this->userLocale).": ".$this->min_contract_term;
		$this->writeHTMLCell(0, 0, 20, $this->GetY(), $billing_point, 0, 1, 0, true);
	}

	private function addItems()
	{
		$columns = array(
			"id"								=> array("name" => "#", "w" => 0.04),
			"line_item_description"				=> array("name" => blmsg("TRANS_DESCRIPTION", array(), $this->userLocale), "w" => 0.43),
			"line_item_quantity"				=> array("name" => blmsg("TRANS_QUANTITY_ABBR", array(), $this->userLocale), "w" => 0.05),
			"line_item_cycle"					=> array("name" => blmsg("TRANS_PDF_INVOICE_CYCLE", array(), $this->userLocale), "w" => 0.07),
			"line_item_price_amount"			=> array("name" => blmsg("TRANS_PRICE", array(), $this->userLocale), "format" => "money", "w" => 0.08),
			"line_item_discount_amount"			=> array("name" => blmsg("TRANS_DISCOUNT", array(), $this->userLocale), "w" => 0.15),
			"line_item_tax_amount"				=> array("name" => blmsg("TRANS_TAX", array(), $this->userLocale), "format" => "money", "w" => 0.08),
			"line_item_total_amount"			=> array("name" => blmsg("TRANS_TOTAL_PRICE", array(), $this->userLocale), "format" => "money", "w" => 0.10),
		);

		$this->addTableHeader($columns, 5);

		$items = $this->invoice->InvoiceLineItemList(array("package_addon_id" => 0));

		global $php_errormsg;
		if (false === usort($items, array($this, "cmp_package_id")))
			throw new Exception("usort() failed: " . $php_errormsg);

		foreach ($items as $invoiceLineItem) {

			$this->fakePdf->setY(0);

			// print element on fake PDF document to estimate it height
			$save_nn = $this->nn;
			$this->addOneItem($this->fakePdf, $invoiceLineItem, $columns);
			$this->nn = $save_nn;

			if (($this->getY() + $this->fakePdf->getY()) > ($this->tm + $this->ph)) {	// element does not fit to page
				$this->AddPage();
				$this->addTableHeader($columns, 5);
			}

			$this->addOneItem($this, $invoiceLineItem, $columns);

			// add addons
			$subscriptionLineItems = $this->invoice->InvoiceLineItemList(array("package_id" => $invoiceLineItem->package_id));
			$addons = array();
			foreach ($subscriptionLineItems as $subscriptionLineItem) {
				if (0 != $subscriptionLineItem->package_addon_id) { // invoice addons stored in the same table as line items with the package_addon_id != 0
					$addons[] = $subscriptionLineItem;
				}
			}
			if (!empty($addons)) {
				$this->addAddons($addons);
			}
		}
	}

	private function addTotals(TCPDF $pdf)
	{
		$pdf->SetFont($this->getDefaultFont(), "B", 9);

		$tax = 0;
		foreach ($this->invoice->InvoiceLineItemList() as $invoiceLineItem)
			$tax += $invoiceLineItem->line_item_tax_amount;

		$tax_included_title = $this->configParams["tax_included_title"];
		$tax_included_title = ("" != $tax_included_title) ? $tax_included_title : blmsg("TRANS_TAX_TOTAL", array(), $this->userLocale);

		$text = "<table>"
			. "<tr><td>" . blmsg("TRANS_SUB_TOTAL", array(), $this->userLocale) . "</td><td>" . $this->formatValue("money", $this->invoice->invoice_sub_total, $this->invoice->currency_id) . "</td></tr>"
			. "<tr><td>" . $tax_included_title . "</td><td>" . $this->formatValue("money", $tax, $this->invoice->currency_id) . "</td></tr>"
			. "<tr><td>" . blmsg("TRANS_INVOICE_TOTAL", array(), $this->userLocale) . "</td><td>" . $this->formatValue("money", $this->invoice->invoice_amount, $this->invoice->currency_id) . "</td></tr>"
			. "</table>";

		$pdf->writeHTMLCell(0, 0, 141, $pdf->GetY(), $text, 0, 1, 0, true);
		$pdf->SetFont($this->getDefaultFont(), "", $this->getDefaultFontSize());
	}

	private function addPayments()
	{
		if (empty($this->transactions)) {
			return;
		}

		$columns = array(
			"payment_method"		=> array("name" => blmsg("TRANS_BILLING_ACCOUNT_TYPE", array(), $this->userLocale), "w" => 0.7),
			"transaction_amount"	=> array("name" => blmsg("TRANS_AMOUNT", array(), $this->userLocale), "format" => "money", "w" => 0.3),
		);

		$this->addTableHeader($columns, 5);


		foreach ($this->transactions as $transaction) {
			$this->fakePdf->setY(0);
			// print element on fake PDF document to estimate it height
			$this->addOnePayment($this->fakePdf, $transaction, $columns);
			if (($this->getY() + $this->fakePdf->getY()) > ($this->tm + $this->ph)) {	// element does not fit to page
				$this->AddPage();
				$this->addTableHeader($columns, 5);
			}
			$this->addOnePayment($this, $transaction, $columns);
		}
		if (!empty($this->transactions)) {
			$this->SetY($this->GetY() + 10);
		}
	}

	private function addOnePayment(TCPDF $pdf, Transaction $transaction, array $columns)
	{
		$this->addLine($pdf);

		$x = $pdf->GetX();
		$y = $pdf->GetY();

		$maxY = 0;
		require_once('objects/ClientBillingAccount.php');
		require_once('objects/PaymentGateway.php');
		$card_types = PaymentGateway::card_types();
		$billing_account_type = is_null($transaction->BillingAccount) ? NULL : $transaction->BillingAccount->billing_account_type;
		foreach ($columns as $col_name => $desc) {
			switch ($col_name) {
			case "payment_method":
				switch ($billing_account_type){
					case NULL:
						$account = "";
						break;
					case ClientBillingAccount::TYPE_CREDIT_CARD :
						$account = blmsg($card_types[$transaction->BillingAccount->billing_card_type], array(), $this->userLocale).": **** **** **** ".$transaction->BillingAccount->billing_account_num_last_four;
						break;
					case ClientBillingAccount::TYPE_BANK_DRAFT :
					case ClientBillingAccount::TYPE_ECHECK :
						$placeholders = array(
				    		"BANK_NAME" => $transaction->BillingAccount->billing_bank_name,
				    		"LAST_FOUR" => $$transaction->BillingAccount->billing_account_num_last_four,
				    	);
						$account = fmsg(ClientBillingAccount::ACCOUNT_ECHECK_NAME_IN_LIST,$placeholders);
						break;
					case ClientBillingAccount::TYPE_HSPC_BANK_TRANSFER :
						$placeholders = array(
				    		"BANK_TRANSFER" => blmsg("TRANS_HSPC_BANK_TRANSFER_ACCOUNT", array(), $this->userLocale),
				    		"NUMBER" => $transaction->BillingAccount->bankTransferPublicNumber,
				    	);
						$account = fmsg(ClientBillingAccount::ACCOUNT_ECHECK_NAME_IN_LIST,$placeholders);
						break;
					default:
						$account = "";
						break;
				}
				$text = "<strong>" . blmsg($this->payment_types[$transaction->transaction_payment_method], array(), $this->userLocale) ." ". $account . "</strong>";
				break;
			case "transaction_amount":
				$text =  $this->formatValue("money", $transaction->transaction_amount, $transaction->currency_id);
				break;
			}

			$pdf->MultiCell($desc["w"] * $this->pw, 0, $text, 0, "L", 0, 0, $x, $y, true, 0, true);
			$x = $pdf->GetX();
			$y = $pdf->GetY();
			$pdf->Ln();
			$y2 = $pdf->GetY();
			$maxY = max($maxY, $y2);
		}

		$pdf->Ln();
		$pdf->SetY($maxY);
	}

	private function addOneItem(TCPDF $pdf, InvoiceLineItem $invoiceLineItem, array $columns)
	{
		$this->addLine($pdf);

		$x = $pdf->GetX();
		$y = $pdf->GetY();

		$maxY = 0;

		foreach ($columns as $col_name => $desc) {

			switch ($col_name) {
			case "id":
				$text = "<strong>" . $this->nn++ . "</strong>";
				break;
			case "line_item_discount_amount":
				global $CONF;
				$text =  $this->formatValue("money", $invoiceLineItem->$col_name, $this->invoice->currency_id);
				if (0 < $invoiceLineItem->line_item_discount_amount) {
					$price = $this->formatValue("money", $invoiceLineItem->line_item_price_amount, $this->invoice->currency_id);
					$text .= "<br />" . blmsg("TRANS_BASIS", array(), $this->userLocale) . ": $price@ ";
					if ((0 < $invoiceLineItem->line_item_discount) && (0 < $invoiceLineItem->proration_discount))
						$text .= ($invoiceLineItem->line_item_discount * 100) . "% + " . ($invoiceLineItem->proration_discount * 100) . "%";
					else if (0 < $invoiceLineItem->line_item_discount)
						$text .= ($invoiceLineItem->line_item_discount * 100) . "%";
					else
						$text .= ($invoiceLineItem->proration_discount * 100) . "%";
				}
				break;
			case "line_item_cycle":
				$text = blmsg($invoiceLineItem->Cycle->cycle_display_name, array(), $this->userLocale);
				break;
			case "line_item_description":
				$text = $invoiceLineItem->$col_name . "<br />"
					. $this->formatValue("date", $invoiceLineItem->line_item_date_begin_renewal) . " - "
					. $this->formatValue("date", $this->getDateMinusDay($invoiceLineItem->line_item_date_next_renewal));

				foreach ($invoiceLineItem->Package->PackageDomainList() as $domain)
					$text .= "<br />" . blmsg("TRANS_DOMAIN", array(), $this->userLocale) . ":&nbsp;<strong>{$domain->domain_sld}.{$domain->domain_tld}</strong>";
				break;
			case "line_item_total_amount":
				$text = $invoiceLineItem->line_item_total_amount;
				break;
			default:
				$text = $invoiceLineItem->$col_name;
				break;
			}

			$text = $this->formatValue($desc["format"], $text, $this->invoice->currency_id);

			$pdf->MultiCell($desc["w"] * $this->pw, 0, $text, 0, "L", 0, 0, $x, $y, true, 0, true);
			$x = $pdf->GetX();
			$y = $pdf->GetY();
			$pdf->Ln();
			$y2 = $pdf->GetY();
			$maxY = max($maxY, $y2);
		}

		$pdf->Ln();
		$pdf->SetY($maxY);
		$x = $pdf->GetX();
		$y = $pdf->GetY();

		if (0 < $invoiceLineItem->line_item_setup_amount) {
			foreach ($columns as $col_name => $desc) {
				switch ($col_name) {
					case "line_item_description":
						$text = blmsg("TRANS_SETUP", array(), $this->userLocale);
						break;
					case "line_item_price_amount":
						$text = $this->formatValue("money", $invoiceLineItem->line_item_setup_amount, $this->invoice->currency_id);
						break;
					default:
						$text = "";
						break;
				}

				$pdf->MultiCell($desc["w"] * $this->pw, 0, $text, 0, "L", 0, 0, $x, $y, true, 0, true);
				$x = $pdf->GetX();
				$y = $pdf->GetY();
				$pdf->Ln();
				$y2 = $pdf->GetY();
				$maxY = max($maxY, $y2);
			}
		}

		$pdf->Ln();
		$pdf->SetY($maxY);

		/* Package comments for invoice are turned off by default. If you want to output it, please, uncomment this part of the code.

		if ("" != $invoiceLineItem->Package->package_comments) {
			$pdf->MultiCell($columns["id"]["w"] * $this->pw, 0, "", 0, "L", 0, 0);
			$this->addLine($pdf);
			$comm = "<font size=\"-1\"><i>" . $invoiceLineItem->Package->package_comments . "</i></font>";
			$pdf->writeHTMLCell(0, 0, $pdf->GetX(), $pdf->GetY(), $comm, 0, 1, 0, true);
		}

		*/
	}

	private function addAddons(array $addons)
	{
		$columns = array(
			"id"						=> array("name" => blmsg("TRANS_ADDONS", array(), $this->userLocale), "w" => 0.08),
			"line_item_description"		=> array("name" => blmsg("TRANS_DESCRIPTION", array(), $this->userLocale), "w" => 0.40),
			"line_item_quantity"		=> array("name" => blmsg("TRANS_QUANTITY_ABBR", array(), $this->userLocale), "w" => 0.05),
			"line_item_cycle"			=> array("name" => blmsg("TRANS_PDF_INVOICE_CYCLE", array(), $this->userLocale), "w" => 0.07),
			"line_item_price_amount"	=> array("name" => blmsg("TRANS_PRICE", array(), $this->userLocale), "format" => "money", "w" => 0.08),
			"line_item_discount_amount"	=> array("name" => blmsg("TRANS_DISCOUNT", array(), $this->userLocale), "w" => 0.14),
			"line_item_tax_amount"		=> array("name" => blmsg("TRANS_TAX", array(), $this->userLocale), "format" => "money", "w" => 0.08),
			"line_item_total_amount"	=> array("name" => blmsg("TRANS_TOTAL_PRICE", array(), $this->userLocale), "format" => "money", "w" => 0.10),
		);

		$this->addTableSubHeader($columns);

		global $php_errormsg;
		if (false === usort($addons, array($this, "cmp_package_addon_id")))
			throw new Exception("usort() failed: " . $php_errormsg);

		foreach ($addons as $addon) {
			$this->fakePdf->setY(0);

			// print element on fake PDF document to estimate it height
			$save_nn = $this->nn;
			$this->addOneAddon($this->fakePdf, $addon, $columns);
			$this->nn = $save_nn;

			if (($this->getY() + $this->fakePdf->getY()) > ($this->tm + $this->ph)) {	// element does not fit to page
				$this->AddPage();
				$this->addTableSubHeader($columns);
			}

			$this->addOneAddon($this, $addon, $columns);
		}
	}

	private function addOneAddon(TCPDF $pdf, InvoiceLineItem $addon, array $columns)
	{

		$x = $pdf->GetX();
		$y = $pdf->GetY();

		$maxY = 0;

		foreach ($columns as $col_name => $desc) {

			switch ($col_name) {
			case "id":
				$text = $this->nn++;
				break;
			case "line_item_cycle":
				$text = blmsg($addon->Cycle->cycle_display_name, array(), $this->userLocale);
				break;
			case "line_item_discount_amount":
				global $CONF;
				$text =  $this->formatValue("money", $addon->$col_name, $this->invoice->currency_id);
				if (0 < $addon->line_item_discount_amount) {
					$price = $this->formatValue("money", $addon->line_item_price_amount, $this->invoice->currency_id);
					$text .= "<br>" . blmsg("TRANS_BASIS", array(), $this->userLocale) . ": $price@ ";
					if ((0 < $addon->line_item_discount) && (0 < $addon->proration_discount))
						$text .= ($addon->line_item_discount * 100) . "% + " . ($addon->proration_discount * 100) . "%";
					else if (0 < $addon->line_item_discount)
						$text .= ($addon->line_item_discount * 100) . "%";
					else
						$text .= ($addon->proration_discount * 100) . "%";
				}
				break;
			case "line_item_total_amount":
				$text = $addon->line_item_total_amount;
				break;
			default:
				$text = $addon->$col_name;
				break;
			}

			$text = $this->formatValue($desc["format"], $text, $this->invoice->currency_id);

			$pdf->MultiCell($desc["w"] * $this->pw, 0, $text, 0, "L", 0, 0, $x, $y, true, 0, true);
			$x = $pdf->GetX();
			$y = $pdf->GetY();
			$pdf->Ln();
			$y2 = $pdf->GetY();
			$maxY = max($maxY, $y2);
		}

		$pdf->Ln();
		$pdf->SetY($maxY);

		$x = $pdf->GetX();
		$y = $pdf->GetY();

		if (0 < $addon->line_item_setup_amount) {
			foreach ($columns as $col_name => $desc) {
				switch ($col_name) {
					case "line_item_description":
						$text = blmsg("TRANS_SETUP", array(), $this->userLocale);
						break;
					case "line_item_price_amount":
						$text = $this->formatValue("money", $addon->line_item_setup_amount, $this->invoice->currency_id);
						break;
					default:
						$text = "";
						break;
				}

				$pdf->MultiCell($desc["w"] * $this->pw, 0, $text, 0, "L", 0, 0, $x, $y, true, 0, true);
				$x = $pdf->GetX();
				$y = $pdf->GetY();
				$pdf->Ln();
				$y2 = $pdf->GetY();
				$maxY = max($maxY, $y2);
			}
		}
	}

	private function addSaleTerms(TCPDF $pdf)
	{
		$sale_terms = $this->configParams["sale_terms"];
		if ("" != $sale_terms)	{
			$pdf->writeHTMLCell(0, 0, $pdf->GetX(), $pdf->GetY(), $sale_terms, 0, 1, 0, true);
			$pdf->SetY($pdf->GetY() + 5);
		}
	}

	private function addCustomFooter(TCPDF $pdf)
	{
		$custom_footer = $this->configParams["custom_footer"];
		if ("" != $custom_footer) {
			$pdf->writeHTMLCell(0, 0, $pdf->GetX(), $pdf->GetY(), $custom_footer, 0, 1, 0, true);
		}
	}

	private function addTableName($name)
	{
		$this->SetFillColor(180);

		$this->SetTextColor(255);
		$this->SetFont($this->getDefaultFont(), "", $this->getDefaultFontSize() + 3);

		$this->MultiCell($this->pw, 0, $name, 0, "L", 1, 1);

		$this->SetTextColor(0);
		$this->SetFont($this->getDefaultFont(), "", $this->getDefaultFontSize());
	}

	private function addTableSubHeader(array $columns)
	{
		$this->addLine($this);
		$this->SetFont($this->getDefaultFont(), "I", $this->getDefaultFontSize());

		foreach ($columns as $desc)
			$this->MultiCell($desc["w"] * $this->pw, 0, $desc["name"], 0, "L", 0, 0);

		$this->SetFont($this->getDefaultFont(), "", $this->getDefaultFontSize());
		$this->Ln();
	}

	private function addTableHeader(array $columns, $minHeight = 0)
	{
		$this->SetFillColor(235);
		$this->SetFont($this->getDefaultFont(), "B", $this->getDefaultFontSize());

		$x = $this->GetX();
		$y = $this->GetY();

		$maxY = 0;
		foreach ($columns as $desc) {
			$this->MultiCell($desc["w"] * $this->pw, $minHeight, $desc["name"], 0, "L", 1, 0, $x, $y, true);
			$x = $this->GetX();
			$y = $this->GetY();
			$this->Ln();
			$y2 = $this->GetY();
			$maxY = max($maxY, $y2);
		}

		$this->SetFont($this->getDefaultFont(), "", $this->getDefaultFontSize());

		$this->Ln();
		$this->SetY($maxY);
	}

	private function addLine(TCPDF $pdf)
	{
		$pdf->SetLineWidth(0.3);
		$pdf->SetDrawColor(180);
		$pdf->Line($pdf->GetX(), $pdf->GetY(), $this->lm + $this->pw, $pdf->GetY());
	}

	private function formatValue($format, $text, $currency_id = "USD")
	{
		switch ($format) {
			case "money":
				return BillingLocale::getCurrency($text, $currency_id, $this->issuerLocale);
			case "date":
				return BillingLocale::getDateMedium($text, $this->issuerLocale);
			default:
				return $text;
		}
	}

	private function initAddressOptions()
	{
		$units = ("mm" == $this->configParams["measurement_unit"]) ? 1 : 25.4;	// 1 - inches, 2 - millimeters

		$w = $this->configParams["horizontal_size"];
		$this->address_w = ("" == $w) ? 114.3 : ($w * $units);

		$h = $this->configParams["vertical_size"];
		$this->address_h = ("" == $h) ? 28.575 : ($h * $units);

		$x = $this->configParams["horizontal_indent"];
		$this->address_x = ("" == $x) ? 22.225 : ($x * $units);

		$y = $this->configParams["vertical_indent"];
		$this->address_y = ("" == $y) ? 57.15 : ($y * $units);

		$fs = $this->configParams["font_size"];
		$this->address_font_size = ("" == $fs) ? 10 : $fs;
	}

	private function cmp_id(DataObject $a, DataObject $b)
	{
		return $a->id - $b->id;
	}

	private function cmp_package_addon_id(DataObject $a, DataObject $b)
	{
		return $a->package_addon_id - $b->package_addon_id;
	}

	private function cmp_package_id(DataObject $a, DataObject $b)
	{
		return $a->package_id - $b->package_id;
	}

	private function getDateMinusDay($text)
	{
		return intval($text)-86400;
	}

	/*
	 * @var Invoice
     */
	private $invoice;

	/*
	 * @var Client
     */
	private $client;

	/*
	 * @var Primary Contact
     */
	private $primaryContact;

	/*
	 * @var Company
     */
	private $issuerCompany;

	/*
	 * Configuration parameters from 'Invoice Settings' page
	 *
	 * @var array
     */
	private $configParams;

	/*
	 * Stationery template index
	 *
	 * @var integer
	 */
	private $stationery;

	/*
	 * Page width without margins
	 *
	 * @var float
	 */
	private $pw;

	/*
	 * Page height without margins
	 *
	 * @var float
	 */
	private $ph;

	/*
	 * Page left margin
	 *
	 * @var float
	 */
	private $lm;

	/*
	 * Page top margin
	 *
	 * @var float
	 */
	private $tm;

	/*
	 * Fake pdf document for height estimations
	 *
	 * @var FPDI
	 */
	private $fakePdf;

	/*
	 * Order number for invoice items
	 *
	 * @var integer
	 */
	private $nn = 1;

	/*
	 * Address blocks top margin
	 *
	 * @var integer
	 */
	private $address_y;

	/*
	 * Address blocks weight
	 *
	 * @var integer
	 */
	private $address_w;

	/*
	 * Address blocks height
	 *
	 * @var integer
	 */
	private $address_h;

	/*
	 * Address blocks left margin
	 *
	 * @var integer
	 */
	private $address_x;

	/*
	 * Address blocks header align (left/right)
	 *
	 * @var integer
	 */
	private $address_header_align;

	/*
	 * Address blocks font size
	 *
	 * @var integer
	 */
	private $address_font_size;

	/*
	 * List of successfull transactions
	 *
	 * @var TransactionList
	 */
	private $transactions;

	/*
	 * List of payment types
	 *
	 * @var array
	 */
	private $payment_types;

	/*
	 * Minimal contract term
	 *
	 * @var string
	 */
	private $min_contract_term = "";

	/*
	 * Issuer locale
	 *
	 * @var string
	 */
	private $issuerLocale;

	/*
	 * Invoice user locale
	 *
	 * @var string
	 */
	private $userLocale;
}
