<?php

require_once(dirname(__FILE__)."/../../../../../lib-tk/shield.php");
require_once("objects/ClientBillingAccount.php");

abstract class HSPC_PM_Adapter
{
	const ACCOUNT_DATA_SEPARATOR = "##";
	const VALIDATION_OK = "OK";

	public function __construct(ClientBillingAccount $account, $gateway_id = NULL)
	{
		$this->gateway_id = is_null($gateway_id) ? $account->payment_gateway_id : $gateway_id;
		$this->account = $account;
	}

	public function setHSPCData(array $hspc_public_data, array $hspc_secure_data)
	{
		foreach ($this->fieldsDescriptor as $account_field => $descriptor) {
			$this->setData($account_field, $this->composeAccountData($descriptor, $hspc_public_data, $hspc_secure_data));
		}
	}

	public function getHSPCData() // must returns map with 'hspc_public_data' and 'hspc_secure_data'
	{
		$secure_data = array();
		$public_data = array();

		foreach ($this->fieldsDescriptor as $account_field => $descriptor) {
			foreach (array_keys($descriptor) as $hspc_field) {
				$data = $this->getHSPCDataByFieldDescriptor($account_field, $descriptor, $hspc_field);
				if (in_array($hspc_field, $this->secureFields)) {
					$secure_data[$hspc_field] = $data;
					continue;
				}
				$public_data[$hspc_field] = $data;
			}
		}
		return array(
			"public_data" => $public_data,
			"secure_data" => $secure_data,
		);
	}

	public function getInputFields($dont_fill_input = false)
	{
		$result = array();
		foreach ($this->fieldsDescriptor as $account_field => $descriptor) {
			foreach ($descriptor as $hspc_field => $hspc_field_opts) {
				$input_field_name = get_class($this)."_".$hspc_field;
				$result[$input_field_name] = array(
					$hspc_field_opts[0],
					$dont_fill_input ? NULL : $this->getHSPCDataByFieldDescriptor($account_field, $descriptor, $hspc_field),
					$hspc_field_opts[1]
				);
			}
		}
		/*
			array(
				"field_name" => array("title", "value", "description"),
				...
			)
		*/
		return $result;
	}

	public function getBillingAccountFields()
	{
		return array_keys($this->fieldsDescriptor);
	}

	public function collectInputFields($suffix = "")
	{
		foreach ($this->fieldsDescriptor as $account_field => $descriptor) {
			$this->setData($account_field, $this->composeAccountDataFromRequest($descriptor, $suffix));
		}
	}

	public function validate()
	{
		$account_data = $this->getHSPCData();
		require_once("HSPC/PaymentMethod.php");
		$hspc_result = HSPCPaymentMethod::validateData($this->getPaymentMethod(), $account_data["secure_data"], $account_data["public_data"]);
		return "" != $hspc_result->data->user_error ? $hspc_result->data->user_error : self::VALIDATION_OK;
	}

	public function save()
	{
		require_once("object_managers/ClientBillingAccountManager.php");
		$this->account->billing_account_num_last_four = $this->getLastFourDigit();
		ClientBillingAccountManager::update($this->account);
	}

	protected function getData($account_field)
	{
		return $this->account->$account_field;
	}

	protected function setData($account_field, $data)
	{
		$this->account->$account_field = $data;
	}

	protected function getLastFourDigit()
	{
		return substr($this->account->billing_bank_account_num, -4);
	}

	protected function getHSPCDataByFieldDescriptor($account_field, $field_descriptor, $hspc_field)
	{
		if ("billing_bank_account_num" == $account_field && NULL === CryptManager::getKey()) {
			return NULL;
		}
		$parsed_data = $this->parseAccountData($field_descriptor, $this->getData($account_field));
		return $parsed_data[$hspc_field];
	}

	protected function parseAccountData($field_descriptor, $data)
	{
		$result = array();
		$parts = split(self::ACCOUNT_DATA_SEPARATOR, $data);
		$parts_idx = 0;

		foreach (array_keys($field_descriptor) as $hspc_field) {
			$result[$hspc_field] = $parts[$parts_idx++];
		}
		return $result;
	}

	protected function composeAccountData($field_descriptor, $hspc_public_data, $hspc_secure_data)
	{
		$data = array();
		foreach (array_keys($field_descriptor) as $hspc_field) {
			$hspc_data = in_array($hspc_field, $this->secureFields) ? $hspc_secure_data : $hspc_public_data;
			$data[] = $hspc_data[$hspc_field];
		}
		return 0 < count($data) ? join(self::ACCOUNT_DATA_SEPARATOR, $data) : NULL;
	}

	protected function composeAccountDataFromRequest($field_descriptor, $suffix = "")
	{
		$data = array();
		foreach (array_keys($field_descriptor) as $hspc_field) {
			$data[] = bgpc(get_class($this)."_".$hspc_field.$suffix);
		}
		return 0 < count($data) ? join(self::ACCOUNT_DATA_SEPARATOR, $data) : NULL;
	}

	protected function getPaymentMethod()
	{
		return preg_replace("/^HSPC_PM_/", "", get_class($this));
	}

	protected $publicDescriptor = array();
	protected $secureDescriptor = array();
	protected $gateway_id;
	protected $account;
}